@@ -1,4 +1,4 @@
-# Copyright (c) 1997-2000 Graham Barr <gbarr@pobox.com>. All rights reserved.
+# Copyright (c) 1997-2004 Graham Barr <gbarr@pobox.com>. All rights reserved.
 # This program is free software; you can redistribute it and/or
 # modify it under the same terms as Perl itself.
 
@@ -7,12 +7,14 @@ package Net::LDAP;
 use strict;
 use IO::Socket;
 use IO::Select;
+use Tie::Hash;
 use vars qw($VERSION $LDAP_VERSION @ISA);
 use Convert::ASN1 qw(asn_read);
 use Net::LDAP::Message;
 use Net::LDAP::ASN qw(LDAPResponse);
 use Net::LDAP::Constant qw(LDAP_SUCCESS
 			   LDAP_OPERATIONS_ERROR
+			   LDAP_SASL_BIND_IN_PROGRESS
 			   LDAP_DECODING_ERROR
 			   LDAP_PROTOCOL_ERROR
 			   LDAP_ENCODING_ERROR
@@ -20,11 +22,15 @@ use Net::LDAP::Constant qw(LDAP_SUCCESS
 			   LDAP_LOCAL_ERROR
 			   LDAP_PARAM_ERROR
 			   LDAP_INAPPROPRIATE_AUTH
+			   LDAP_SERVER_DOWN
+			   LDAP_USER_CANCELED
+			   LDAP_EXTENSION_START_TLS
+			   LDAP_UNAVAILABLE
 			);
 
-$VERSION 	= "0.26";
-@ISA     	= qw(Net::LDAP::Extra);
-$LDAP_VERSION 	= 2;      # default LDAP protocol version
+$VERSION 	= "0.4001";
+@ISA     	= qw(Tie::StdHash Net::LDAP::Extra);
+$LDAP_VERSION 	= 3;      # default LDAP protocol version
 
 # Net::LDAP::Extra will only exist is someone use's the module. But we need
 # to ensure the package stash exists or perl will complain that we inherit
@@ -96,19 +102,25 @@ sub new {
   my $arg  = &_options;
   my $obj  = bless {}, $type;
 
-  foreach my $h (ref($host) ? @$host : ($host)) {
-    if ($obj->_connect($host, $arg)) {
-      $obj->{net_ldap_host} = $h;
+  foreach my $uri (ref($host) ? @$host : ($host)) {
+    my $scheme = $arg->{scheme} || 'ldap';
+    (my $h = $uri) =~ s,^(\w+)://,, and $scheme = $1;
+    my $meth = $obj->can("connect_$scheme") or next;
+    $h =~ s,/.*,,; # remove path part
+    $h =~ s/%([A-Fa-f0-9]{2})/chr(hex($1))/eg; # unescape
+    if (&$meth($obj, $h, $arg)) {
+      $obj->{net_ldap_uri} = $uri;
+      $obj->{net_ldap_scheme} = $scheme;
       last;
     }
   }
 
   return undef unless $obj->{net_ldap_socket};
 
-  $obj->{net_ldap_host}    = $host;
   $obj->{net_ldap_resp}    = {};
   $obj->{net_ldap_version} = $arg->{version} || $LDAP_VERSION;
   $obj->{net_ldap_async}   = $arg->{async} ? 1 : 0;
+  $obj->{raw} = $arg->{raw}  if ($arg->{raw});
 
   if (defined(my $onerr = $arg->{onerror})) {
     $onerr = $onerror{$onerr} if exists $onerror{$onerr};
@@ -117,20 +129,127 @@ sub new {
 
   $obj->debug($arg->{debug} || 0 );
 
-  $obj;
+  $obj->outer;
 }
 
-sub _connect {
+sub connect_ldap {
   my ($ldap, $host, $arg) = @_;
+  my $port = $arg->{port} || 389;
+  my $class = 'IO::Socket::INET';
 
-  $ldap->{net_ldap_socket} = IO::Socket::INET->new(
-    PeerAddr => $host,
-    PeerPort => $arg->{port} || '389',
-    Proto    => 'tcp',
+  # separate port from host overwriting given/default port
+  $host =~ s/^([^:]+|\[.*\]):(\d+)$/$1/ and $port = $2;
+
+  if ($arg->{inet6}) {
+    require IO::Socket::INET6;
+    $class = 'IO::Socket::INET6';
+  }  
+
+  $ldap->{net_ldap_socket} = $class->new(
+    PeerAddr   => $host,
+    PeerPort   => $port,
+    LocalAddr  => $arg->{localaddr} || undef,
+    Proto      => 'tcp',
+    MultiHomed => $arg->{multihomed},
+    Timeout    => defined $arg->{timeout}
+		 ? $arg->{timeout}
+		 : 120
+  ) or return undef;
+  
+  $ldap->{net_ldap_host} = $host;
+  $ldap->{net_ldap_port} = $port;
+}
+
+
+# Different OpenSSL verify modes.
+my %ssl_verify = qw(none 0 optional 1 require 3);
+
+sub connect_ldaps {
+  my ($ldap, $host, $arg) = @_;
+  my $port = $arg->{port} || 636;
+
+  require IO::Socket::INET6  if ($arg->{inet6});
+  require IO::Socket::SSL;
+  IO::Socket::SSL->import(qw/inet6/)  if ($arg->{inet6});
+
+  # separate port from host overwriting given/default port
+  $host =~ s/^([^:]+|\[.*\]):(\d+)$/$1/ and $port = $2;
+
+  $ldap->{'net_ldap_socket'} = IO::Socket::SSL->new(
+    PeerAddr 	    => $host,
+    PeerPort 	    => $port,
+    LocalAddr       => $arg->{localaddr} || undef,
+    Proto    	    => 'tcp',
+    Timeout  	    => defined $arg->{'timeout'} ? $arg->{'timeout'} : 120,
+    _SSL_context_init_args($arg)
+  ) or return undef;
+
+  $ldap->{net_ldap_host} = $host;
+  $ldap->{net_ldap_port} = $port;
+}
+
+sub _SSL_context_init_args {
+  my $arg = shift;
+
+  my $verify = 0;
+  my ($clientcert,$clientkey,$passwdcb);
+
+  if (exists $arg->{'verify'}) {
+      my $v = lc $arg->{'verify'};
+      $verify = 0 + (exists $ssl_verify{$v} ? $ssl_verify{$v} : $verify);
+  }
+
+  if (exists $arg->{'clientcert'}) {
+      $clientcert = $arg->{'clientcert'};
+      if (exists $arg->{'clientkey'}) {
+	  $clientkey = $arg->{'clientkey'};
+      } else {
+	  require Carp;
+	  Carp::croak("Setting client public key but not client private key");
+      }
+  }
+
+  if ($arg->{'checkcrl'} && !$arg->{'capath'}) {
+      require Carp;
+      Carp::croak("Cannot check CRL without having CA certificates");
+  }
+
+  if (exists $arg->{'keydecrypt'}) {
+      $passwdcb = $arg->{'keydecrypt'};
+  }
+
+  (
+    SSL_cipher_list => defined $arg->{'ciphers'} ? $arg->{'ciphers'} : 'ALL',
+    SSL_ca_file     => exists  $arg->{'cafile'}  ? $arg->{'cafile'}  : '',
+    SSL_ca_path     => exists  $arg->{'capath'}  ? $arg->{'capath'}  : '',
+    SSL_key_file    => $clientcert ? $clientkey : undef,
+    SSL_passwd_cb   => $passwdcb,
+    SSL_check_crl   => $arg->{'checkcrl'} ? 1 : 0,
+    SSL_use_cert    => $clientcert ? 1 : 0,
+    SSL_cert_file   => $clientcert,
+    SSL_verify_mode => $verify,
+    SSL_version     => defined $arg->{'sslversion'} ? $arg->{'sslversion'} :
+                       'sslv2/3',
+  );
+}
+
+sub connect_ldapi {
+  my ($ldap, $peer, $arg) = @_;
+
+  $peer = $ENV{LDAPI_SOCK} || "/var/run/ldapi"
+    unless length $peer;
+
+  require IO::Socket::UNIX;
+
+  $ldap->{net_ldap_socket} = IO::Socket::UNIX->new(
+    Peer => $peer,
     Timeout  => defined $arg->{timeout}
 		 ? $arg->{timeout}
 		 : 120
-  );
+  ) or return undef;
+
+  $ldap->{net_ldap_host} = 'localhost';
+  $ldap->{net_ldap_peer} = $peer;
 }
 
 sub message {
@@ -160,6 +279,25 @@ sub socket {
   $_[0]->{net_ldap_socket};
 }
 
+sub host {
+  my $ldap = shift;
+  ($ldap->{net_ldap_scheme} ne 'ldapi')
+  ? $ldap->{net_ldap_host}
+  : $ldap->{net_ldap_peer};
+}
+
+sub port {
+  $_[0]->{net_ldap_port} || undef;
+}
+
+sub scheme {
+  $_[0]->{net_ldap_scheme};
+}
+
+sub uri {
+  $_[0]->{net_ldap_uri};
+}
+
 
 sub unbind {
   my $ldap = shift;
@@ -239,11 +377,27 @@ sub bind {
       if $ldap->{net_ldap_version} < 3;
 
     my $sasl = $passwd;
-    # Tell the SASL object our user identifier
-    $sasl->callback( user => "dn: $stash{name}")
-      unless $sasl->callback('user');
+    my $sasl_conn;
 
-    my $sasl_conn = $sasl->client_new("ldap",$ldap->{net_ldap_host});
+    if (ref($sasl) and $sasl->isa('Authen::SASL')) {
+
+      # If we're talking to a round-robin, the canonical name of
+      # the host we are talking to might not match the name we
+      # requested
+      my $connected_name = $ldap->{net_ldap_socket}->peerhost;
+      $connected_name ||= $ldap->{net_ldap_host};
+
+      $sasl_conn = eval {
+        local ($SIG{__DIE__});
+        $sasl->client_new("ldap", $connected_name);
+      };
+    }
+    else {
+      $sasl_conn = $sasl;
+    }
+
+    return _error($ldap, $mesg, LDAP_LOCAL_ERROR, "$@")
+      unless defined($sasl_conn);
 
     # Tell SASL the local and server IP addresses
     $sasl_conn->property(
@@ -253,9 +407,12 @@ sub bind {
 
     my $initial = $sasl_conn->client_start;
 
+    return _error($ldap, $mesg, LDAP_LOCAL_ERROR, $sasl_conn->error)
+      unless defined($initial);
+
     $passwd = {
       mechanism   => $sasl_conn->mechanism,
-      credentials => $initial
+      credentials => $initial,
     };
 
     # Save data, we will need it later
@@ -282,6 +439,9 @@ sub search {
 
   require Net::LDAP::Search;
 
+  $arg->{raw} = $ldap->{raw}
+    if ($ldap->{raw} && !defined($arg->{raw}));
+
   my $mesg = $ldap->message('Net::LDAP::Search' => $arg);
 
   my $control = $arg->{control}
@@ -358,7 +518,7 @@ sub add {
 }
 
 
-my %opcode = ( 'add' => 0, 'delete' => 1, 'replace' => 2);
+my %opcode = ( 'add' => 0, 'delete' => 1, 'replace' => 2, 'increment' => 3);
 
 sub modify {
   my $ldap = shift;
@@ -402,7 +562,7 @@ sub modify {
     }
   }
   else {
-    foreach $op (qw(add delete replace)) {
+    foreach $op (qw(add delete replace increment)) {
       next unless exists $arg->{$op};
       my $opt = $arg->{$op};
       my $opcode = $opcode{$op};
@@ -428,7 +588,7 @@ sub modify {
 	    operation => $opcode,
 	    modification => {
 	      type => $attr,
-	      vals => $val
+	      vals => ref($val) ? $val : [$val]
 	    }
 	  };
 	}
@@ -600,14 +760,21 @@ sub sync {
   my $table = $ldap->{net_ldap_mesg};
   my $err   = LDAP_SUCCESS;
 
+  return $err unless defined $table;
+
   $mid = $mid->mesg_id if ref($mid);
   while (defined($mid) ? exists $table->{$mid} : %$table) {
-    last if $err = $ldap->_recvresp($mid);
+    last if $err = $ldap->process($mid);
   }
 
   $err;
 }
 
+sub disconnect {
+  my $self = shift;
+  _drop_conn($self, LDAP_USER_CANCELED, "Explicit disconnect");
+}
+
 sub _sendmesg {
   my $ldap = shift;
   my $mesg = shift;
@@ -624,8 +791,20 @@ sub _sendmesg {
       if $debug & 4;
   }
 
-  syswrite($ldap->socket, $mesg->pdu, length($mesg->pdu))
-    or return _error($ldap, $mesg, LDAP_LOCAL_ERROR,"$!");
+  my $socket = $ldap->socket
+      or return _error($ldap, $mesg, LDAP_SERVER_DOWN, "$!");
+
+  # send packets in sizes that IO::Socket::SSL can chew
+  # originally it was:
+  #syswrite($socket, $mesg->pdu, length($mesg->pdu))
+  #  or return _error($ldap, $mesg, LDAP_LOCAL_ERROR,"$!")
+  my $to_send = \( $mesg->pdu );
+  my $offset = 0;
+  while($offset < length($$to_send)) {
+    my $n = syswrite($socket, substr($$to_send, $offset, 15000), 15000)
+      or return _error($ldap, $mesg, LDAP_LOCAL_ERROR,"$!");
+    $offset += $n;
+  }
 
   # for CLDAP, here we need to recode when we were sent
   # so that we can perform timeouts and resends
@@ -648,17 +827,17 @@ sub _sendmesg {
     : $mesg;
 }
 
-sub _recvresp {
+sub process {
   my $ldap = shift;
   my $what = shift;
-  my $sock = $ldap->socket;
+  my $sock = $ldap->socket or return LDAP_SERVER_DOWN;
   my $sel = IO::Select->new($sock);
   my $ready;
 
   for( $ready = 1 ; $ready ; $ready = $sel->can_read(0)) {
     my $pdu;
     asn_read($sock, $pdu)
-      or return LDAP_OPERATIONS_ERROR;
+      or return _drop_conn($ldap, LDAP_OPERATIONS_ERROR, "Communications Error");
 
     my $debug;
     if ($debug = $ldap->debug) {
@@ -675,13 +854,20 @@ sub _recvresp {
     my $result = $LDAPResponse->decode($pdu)
       or return LDAP_DECODING_ERROR;
 
-    my $mid = $result->{messageID};
+    my $mid  = $result->{messageID};
+    my $mesg = $ldap->{net_ldap_mesg}->{$mid};
 
-    my $mesg = $ldap->{net_ldap_mesg}->{$mid} or
-      do {
-	print STDERR "Unexpected PDU, ignored\n" if $debug & 10;
-	next;
-      };
+    unless ($mesg) {
+      if (my $ext = $result->{protocolOp}{extendedResp}) {
+	if (($ext->{responseName} || '') eq '1.3.6.1.4.1.1466.20036') {
+	  # notice of disconnection
+	  return _drop_conn($ldap, LDAP_SERVER_DOWN, "Notice of Disconnection");
+	}
+      }
+
+      print STDERR "Unexpected PDU, ignored\n" if $debug & 10;
+      next;
+    }
 
     $mesg->decode($result) or
       return $mesg->code;
@@ -695,6 +881,24 @@ sub _recvresp {
   return LDAP_SUCCESS;
 }
 
+*_recvresp = \&process; # compat
+
+sub _drop_conn {
+  my ($self, $err, $etxt) = @_;
+
+  my $sock = delete $self->{net_ldap_socket};
+  close($sock) if $sock;
+
+  if (my $msgs = delete $self->{net_ldap_mesg}) {
+    foreach my $mesg (values %$msgs) {
+      $mesg->set_error($err, $etxt);
+    }
+  }
+
+  $err;
+}
+
+
 sub _forgetmesg {
   my $ldap = shift;
   my $mesg = shift;
@@ -748,6 +952,7 @@ sub schema {
 		dITContentRules
 		nameForms
 		ldapSyntaxes
+                extendedAttributeInfo
               )],
   );
 
@@ -756,6 +961,7 @@ sub schema {
     : Net::LDAP::Schema->new($mesg->entry);
 }
 
+
 sub root_dse {
   my $ldap = shift;
   my %arg  = @_;
@@ -765,9 +971,15 @@ sub root_dse {
 		  altServer
 		  supportedExtension
 		  supportedControl
+		  supportedFeatures
 		  supportedSASLMechanisms
 		  supportedLDAPVersion
+		  vendorName
+		  vendorVersion
 		)];
+  my $root = $arg{attrs} && $ldap->{net_ldap_root_dse};
+
+  return $root if $root;
 
   my $mesg = $ldap->search(
     base   => '',
@@ -776,7 +988,13 @@ sub root_dse {
     attrs  => $attrs,
   );
 
-  $mesg->entry;
+  require Net::LDAP::RootDSE;
+  $root = $mesg->entry;
+  bless $root, 'Net::LDAP::RootDSE' if $root; # Naughty, but there you go :-)
+
+  $ldap->{net_ldap_root_dse} = $root unless $arg{attrs};
+
+  return $root;
 }
 
 sub start_tls {
@@ -784,6 +1002,7 @@ sub start_tls {
   my $arg  = &_options;
   my $sock = $ldap->socket;
 
+  require IO::Socket::SSL;
   require Net::LDAP::Extension;
   my $mesg = $ldap->message('Net::LDAP::Extension' => $arg);
 
@@ -795,7 +1014,7 @@ sub start_tls {
 
   $mesg->encode(
     extendedReq => {
-      requestName => "1.3.6.1.4.1.1466.20037",
+      requestName => LDAP_EXTENSION_START_TLS,
     }
   );
 
@@ -805,12 +1024,23 @@ sub start_tls {
   return $mesg
     if $mesg->code;
 
-  require Net::LDAPS;
+  delete $ldap->{net_ldap_root_dse};
+
   $arg->{sslversion} = 'tlsv1' unless defined $arg->{sslversion};
-  IO::Socket::SSL::context_init( { Net::LDAPS::SSL_context_init_args($arg) } );
-  IO::Socket::SSL::socketToSSL($sock, {Net::LDAPS::SSL_context_init_args($arg)})
-    ? $mesg
-    : _error($ldap, $mesg, LDAP_OPERATIONS_ERROR, $@);
+  IO::Socket::SSL::context_init( { _SSL_context_init_args($arg) } );
+  my $sock_class = ref($sock);
+
+  return $mesg
+    if IO::Socket::SSL->start_SSL($sock, {_SSL_context_init_args($arg)});
+
+  my $err = $@ || $IO::Socket::SSL::SSL_ERROR || $IO::Socket::SSL::SSL_ERROR || ''; # avoid use on once warning
+
+  if ($sock_class ne ref($sock)) {
+    $err = $sock->errstr;
+    bless $sock, $sock_class;
+  }
+
+  _error($ldap, $mesg, LDAP_OPERATIONS_ERROR, $err);
 }
 
 sub cipher {
@@ -836,5 +1066,29 @@ sub version {
     : $ldap->{net_ldap_version};
 }
 
+sub outer {
+  my $self = shift;
+  return $self if tied(%$self);
+  my %outer;
+  tie %outer, ref($self), $self;
+  ++$self->{net_ldap_refcnt};
+  bless \%outer, ref($self);
+}
+
+sub inner {
+  tied(%{$_[0]}) || $_[0];
+}
+
+sub TIEHASH {
+  $_[1];
+}
+
+sub DESTROY {
+  my $ldap = shift;
+  my $inner = tied(%$ldap) or return;
+  _drop_conn($inner, LDAP_UNAVAILABLE, "Implicit disconnect")
+    unless --$inner->{net_ldap_refcnt};
+}
+
 1;
 
